#include "StdInc.hpp"
#include "FrameBuffer.hpp"



FrameBuffer::FrameBuffer()
    : m_hDibSection(NULL)
    , m_pBits(0)
{
}

FrameBuffer::~FrameBuffer()
{
    Destroy();
}

bool FrameBuffer::Destroy(bool result)
{
    if (m_hDibSection) ::DeleteObject(m_hDibSection);

    m_hDibSection = NULL;
    m_pBits       = NULL;
    return result;
}

bool FrameBuffer::Create(int width, int height)
{
    Destroy();

    ::memset(&m_bmi, 0, sizeof(m_bmi));
    m_bmi.bmiHeader.biSize            = sizeof(m_bmi.bmiHeader);
    m_bmi.bmiHeader.biBitCount        = 32;
    m_bmi.bmiHeader.biPlanes          = 1;
    m_bmi.bmiHeader.biCompression     = BI_RGB;
    m_bmi.bmiHeader.biHeight          = -height;
    m_bmi.bmiHeader.biWidth           = width;
    m_bmi.bmiHeader.biSizeImage       = 0;
    m_bmi.bmiHeader.biClrImportant    = 0;
    m_bmi.bmiHeader.biClrUsed         = 0;
    m_bmi.bmiHeader.biXPelsPerMeter   = 0;
    m_bmi.bmiHeader.biYPelsPerMeter   = 0;

    m_hDibSection = ::CreateDIBSection(NULL, &m_bmi, DIB_RGB_COLORS, (void **)&m_pBits, NULL, 0);
    if (!m_hDibSection) return Destroy(false);

    m_width  = width;
    m_height = height;

    DIBSECTION ds;
    if (!::GetObject(m_hDibSection, sizeof(ds), &ds)) return Destroy(false);
    m_pitch = ds.dsBm.bmWidthBytes;

    m_hDC = ::CreateCompatibleDC(0);
    if (!m_hDC) return Destroy(false);

    SelectObject(m_hDC, m_hDibSection);
    SelectObject(m_hDC, ::GetStockObject(HOLLOW_BRUSH));

//    SetViewportOrgEx(m_hDC, width / 2, height / 2, NULL);

    Clear(0x000000FF);

    return true;
}

void FrameBuffer::BlitToWindow(HDC hDstDC, int dstWidth, int dstHeight)
{
    if (m_hDibSection == NULL) return;
    if (hDstDC == NULL) return;

    ::StretchDIBits(hDstDC, 0, 0, dstWidth, dstHeight, 0, 0, m_width, m_height, m_pBits, &m_bmi, DIB_RGB_COLORS, SRCCOPY);
}

void FrameBuffer::Clear(unsigned int color)
{
    for (int y = 0; y < m_height; y++)
    {
        unsigned long * p = (unsigned long *)(&m_pBits[0 + y * m_pitch]);

        for (int x = 0; x < m_width; x++)
        {
            p[x] = color;
        }
    }
}

void FrameBuffer::HLine(int y)
{
    for (int x = 0; x < m_width; x++)
    {
        *(unsigned long *)(&m_pBits[x * 4 + y * m_pitch]) = 0xffffff00;
    }    
}

void FrameBuffer::VLine(int x)
{
    for (int y = 0; y < m_height; y++)
    {
        *(unsigned long *)(&m_pBits[x * 4 + y * m_pitch]) = 0xffffff00;
    }    
}

void FrameBuffer::Fill(int x1, int y1, int x2, int y2)
{
    for (int y = y1; y <= y2; y++)
    {
        unsigned long * p = (unsigned long *)(&m_pBits[0 + y * m_pitch]);

        for (int x = x1; x <= x2; x++)
        {
            p[x] = 0xff000000;
        }
    }
}

void FrameBuffer::FillXor(double xOffset, double yOffset)
{
    int ofsX = (int)xOffset;
    int ofsY = (int)yOffset;

    for (int y = 0; y < m_height; y++)
    {
        unsigned long * p = (unsigned long *)(&m_pBits[0 + y * m_pitch]);

        for (int x = 0; x < m_width; x++)
        {
            int fx = x + ofsX;
            int fy = y + ofsY;

            int r = (fx ^ fy) & 63;
            int g = (fx ^ fy) & 255;
            int b = (fx ^ fy) & 127;
            p[x] = r | (g << 8) | (b << 16);
        }
    }
}

COLORREF FrameBuffer::Color(unsigned int color)
{
    return RGB(
        ((color >> 16) & 0xFF),
        ((color >>  8) & 0xFF),
        ((color >>  0) & 0xFF));
}
