#include "StdInc.hpp"
#include "FrameBuffer.hpp"

#ifndef HID_USAGE_PAGE_GENERIC
    #define HID_USAGE_PAGE_GENERIC         ((USHORT) 0x01)
#endif

#ifndef HID_USAGE_GENERIC_MOUSE
    #define HID_USAGE_GENERIC_MOUSE        ((USHORT) 0x02)
#endif

FrameBuffer         g_frameBuffer;
double              g_scrollOffsetX = 0.0;
double              g_scrollOffsetY = 0.0;

void OnWmInput(WPARAM wParam, LPARAM lParam);

LRESULT CALLBACK WindowProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch (message)
    {
        case WM_PAINT:
        {
            RECT        clientRect;
            PAINTSTRUCT ps;
            HDC         hDC;

            ::GetClientRect(hWnd, &clientRect);

            hDC = ::BeginPaint(hWnd, &ps);
            if (hDC)
            {
                g_frameBuffer.BlitToWindow(hDC, clientRect.right - clientRect.left, clientRect.bottom - clientRect.top);
                ::EndPaint(hWnd, &ps);
            }

            return 0;
        }

        case WM_KEYDOWN:
            if (wParam == VK_ESCAPE) ::SendMessage(hWnd, WM_CLOSE, 0, 0);
            return 0;

        case WM_DESTROY:
            ::PostQuitMessage(0);
            return 0;

        case WM_INPUT:
            OnWmInput(wParam, lParam);
            return 0;
    }
    return ::DefWindowProc(hWnd, message, wParam, lParam);
}

void OnWmInput(WPARAM wParam, LPARAM lParam)
{
if ((GetAsyncKeyState(VK_LBUTTON) & 0x8000) == 0)
    return;

    HRAWINPUT rawInputHandle = (HRAWINPUT)lParam;

    // Determine the size of the raw input buffer
    UINT rawInputBufferSize = 0;

    UINT numCopiedBytes = ::GetRawInputData(rawInputHandle, RID_INPUT, NULL, &rawInputBufferSize, sizeof(RAWINPUTHEADER));
    if (numCopiedBytes != 0) 
        return;

    // Resize raw input buffer if needed
    BYTE * pRawDataBuffer = new BYTE[rawInputBufferSize];

    // Fill the buffer with raw input data
    numCopiedBytes = ::GetRawInputData(rawInputHandle, RID_INPUT, pRawDataBuffer, &rawInputBufferSize, sizeof(RAWINPUTHEADER));
    if (numCopiedBytes == UINT(-1))
        return ;

    // Dispatch the received raw input
    RAWINPUT * pRawInput = (RAWINPUT *)pRawDataBuffer;

    const RAWMOUSE & rRawMouse = pRawInput->data.mouse;

    // Determine relative mouse motion
    if ((rRawMouse.usFlags & 1) == MOUSE_MOVE_RELATIVE)
    {
        if (rRawMouse.lLastX != 0 || rRawMouse.lLastY != 0)
        {
            ////////////////////////////////////////////////////////////////////////////////////////
            ////////////////////////////////////////////////////////////////////////////////////////
            ////////////////////////////////////////////////////////////////////////////////////////
            ////////////////////////////////////////////////////////////////////////////////////////
            // HIER IST DIE RECHNUNG
            ////////////////////////////////////////////////////////////////////////////////////////
            ////////////////////////////////////////////////////////////////////////////////////////
            ////////////////////////////////////////////////////////////////////////////////////////
            ////////////////////////////////////////////////////////////////////////////////////////
            int nativeMouseSensitivity = 0;
            ::SystemParametersInfo(SPI_GETMOUSESPEED, 0, &nativeMouseSensitivity, 0);

            double d = static_cast<double>(nativeMouseSensitivity) / 10.0;

            g_scrollOffsetX -= static_cast<double>(rRawMouse.lLastX) * d;
            g_scrollOffsetY -= static_cast<double>(rRawMouse.lLastY) * d;
        }
    }

    delete [] pRawDataBuffer;
}

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShowCmd)
{
    WNDCLASSEX  wcex;

    ::memset(&wcex, 0, sizeof(wcex));
    wcex.cbSize         = sizeof(wcex);
    wcex.style          = CS_HREDRAW | CS_VREDRAW;
    wcex.cbClsExtra     = 0;
    wcex.cbWndExtra     = 0;
    wcex.hbrBackground  = (HBRUSH)(::GetStockObject(BLACK_BRUSH));
    wcex.hCursor        = ::LoadCursor(NULL, IDC_ARROW);
    wcex.hIcon          = ::LoadIcon(NULL, IDI_APPLICATION);
    wcex.hIconSm        = ::LoadIcon(NULL, IDI_APPLICATION);
    wcex.lpfnWndProc    = WindowProc;
    wcex.hInstance      = ::GetModuleHandle(NULL);
    wcex.lpszClassName  = TEXT("SensitivityTest_Window_Class");
    wcex.lpszMenuName   = NULL;
    if (!::RegisterClassEx(&wcex)) return -1;

    int scrW = 1024;
    int scrH = 768;

    RECT    rc = { 0, 0, scrW, scrH };
    DWORD   style = WS_OVERLAPPED | WS_SYSMENU | WS_CAPTION | WS_DLGFRAME;

    if (!g_frameBuffer.Create(rc.right, rc.bottom)) return -1;

    AdjustWindowRect(&rc, style, FALSE);

    rc.right  -= rc.left;
    rc.bottom -= rc.top;
    rc.left    = (::GetSystemMetrics(SM_CXSCREEN) - rc.right) / 2;
    rc.top     = (::GetSystemMetrics(SM_CYSCREEN) - rc.bottom) / 2;

    HWND hWnd = ::CreateWindow(
        wcex.lpszClassName,
        TEXT("Sensitivity Test"),
        style,
        rc.left,
        rc.top,
        rc.right,
        rc.bottom,
        NULL,
        NULL,
        ::GetModuleHandle(NULL),
        NULL);
    if (!hWnd) return -1;

    RAWINPUTDEVICE rawInputDevices[1];

    rawInputDevices[0].usUsagePage = HID_USAGE_PAGE_GENERIC; 
    rawInputDevices[0].usUsage     = HID_USAGE_GENERIC_MOUSE; 
    rawInputDevices[0].dwFlags     = 0;//RIDEV_INPUTSINK;   
    rawInputDevices[0].hwndTarget  = hWnd;
    
    ::RegisterRawInputDevices(rawInputDevices, 1, sizeof(rawInputDevices[0]));
    ::ShowWindow(hWnd, SW_SHOW);
    ::UpdateWindow(hWnd);

    MSG msg;


    DWORD   lastTime = ::timeGetTime();

    g_frameBuffer.Clear(0xff112233);

    while (true)
    {
        if (::PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE))
        {
            if (!::GetMessage(&msg, NULL, 0, 0)) return 0;

            ::TranslateMessage(&msg);
            ::DispatchMessage(&msg);
        }
        else
        {
            DWORD curTime = ::timeGetTime();
            DWORD elapsedTime = curTime - lastTime;
            lastTime = curTime;

            double elapsedSeconds = double(elapsedTime) / 1000.0;

            g_frameBuffer.FillXor(g_scrollOffsetX, g_scrollOffsetY);

//g_scrollOffsetX -= elapsedSeconds * 100.0;
//g_scrollOffsetY -= elapsedSeconds * 100.0;

//            g_frameBuffer.Fill(clipD, clipD, g_frameBuffer.m_width-clipD-1, g_frameBuffer.m_height-clipD-1);

            ::InvalidateRect(hWnd, NULL, FALSE);
            ::UpdateWindow(hWnd);
        }
    }

    return 0;
}   
